from pyomo.environ import Suffix, Var, Expression, Constraint
from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.network import Arc, SequentialDecomposition
from pyomo.core.base.reference import Reference
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.math import smooth_max
from idaes.models.unit_models.separator import SplittingType, EnergySplittingType
from idaes.models.unit_models.mixer import Mixer
from idaes.models.unit_models.heater import Heater
from ahuora_builder.custom.custom_separator import CustomSeparator
from ahuora_builder.custom.simple_separator import SimpleSeparator
from ..inverted import add_inverted, initialise_inverted
__author__ = "Team Ahuora"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("SteamHeader")
class SteamHeaderData(UnitModelBlockData):
    """
    Steam Header unit operation:
    Mixer -> Cooler -> Phase Separator -> Simple Separator
    Separates 100% liquid to condensate_outlet and 100% vapor to splitter outlets.
    Uses Sequential Decomposition for optimal initialization order.
    """

    CONFIG = UnitModelBlockData.CONFIG()
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
        ),
    )
    CONFIG.declare(
        "num_inlets",
        ConfigValue(
            default=2,
            domain=int,
            description="Number of inlets to add" "Index [-1]: Steam makeup",
        ),
    )
    CONFIG.declare(
        "num_outlets",
        ConfigValue(
            default=2,
            domain=int,
            description="Number of outlets to add"
            "Index [-1]: Vent"
            "Index [-2]: Condensate",
        ),
    )

    def build(self):
        super().build()
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        self.inlet_list = [f"inlet_{i+1}" for i in range(self.config.num_inlets)]
        self.outlet_list = [
            f"outlet_{i+1}" for i in range(self.config.num_outlets)
        ] + ["vent"]

        # Create internal units
        self.mixer = Mixer(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            num_inlets=self.config.num_inlets,
            inlet_list=self.inlet_list,
        )
        self.cooler = Heater(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            has_pressure_change=True,
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
        )
        self.phase_separator = CustomSeparator(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            outlet_list=["vapor_outlet", "condensate_outlet"],
            split_basis=SplittingType.phaseFlow,
            energy_split_basis=EnergySplittingType.enthalpy_split,
        )
        self.splitter = SimpleSeparator(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            outlet_list=self.outlet_list,
        )
        self.unit_ops = [self.mixer, self.cooler, self.phase_separator, self.splitter]

        # Updated internal arcs
        self.mixer_to_cooler_arc = Arc(
            source=self.mixer.outlet, destination=self.cooler.inlet
        )
        self.cooler_to_separator_arc = Arc(
            source=self.cooler.outlet, destination=self.phase_separator.inlet
        )
        self.separator_to_splitter_arc = Arc(
            source=self.phase_separator.vapor_outlet, destination=self.splitter.inlet
        )

        self.inlet_blocks = [getattr(self.mixer, f"{i}_state") for i in self.inlet_list]
        self.outlet_blocks = [getattr(self.splitter, f"{o}_state") for o in self.outlet_list]

        # Declare slack variables for internal use
        self.balance_flow_mol = Var(
            self.flowsheet().time, initialize=0.0, doc="Balance molar flow (negative means makeup is required, positive means venting)", 
            units=pyo.units.mol / pyo.units.s
        )
        
        # This is only used for scaling the smooth_min/smooth_max value.
        self._partial_total_flow_mol = Var(
            self.flowsheet().time, initialize=0.0, doc="Partial total fixed molar flow",
            units=pyo.units.mol / pyo.units.s
        )

        self.makeup_flow_mol = Var(
            self.flowsheet().time, initialize=0.0, doc="Makeup molar flow",
            units=pyo.units.mol / pyo.units.s
        )


        # Add inverted transformers to heat_duty and deltaP
        # (so that positive values correspond to heat loss and pressure drop)

        add_inverted(self.cooler, "heat_duty")
        add_inverted(self.cooler, "deltaP")
        # Declare additional variables and aliases to expose to the user
        self.heat_duty_inverted = Reference(self.cooler.heat_duty_inverted)
        self.deltaP_inverted = Reference(self.cooler.deltaP_inverted)
        self.heat_duty = Reference(self.cooler.heat_duty)
        self.deltaP = Reference(self.cooler.deltaP)
        
        self.total_flow_mol = Reference(
            self.cooler.control_volume.properties_out[:].flow_mol
        )
        self.total_flow_mass = Reference(
            self.cooler.control_volume.properties_out[:].flow_mass
        )
        self.pressure = Reference(
            self.cooler.control_volume.properties_out[:].pressure
            )
        self.temperature = Reference(
            self.cooler.control_volume.properties_out[:].temperature
        )
        self.enth_mol = Reference(
            self.cooler.control_volume.properties_out[:].enth_mol
            )
        self.enth_mass = Reference(
            self.cooler.control_volume.properties_out[:].enth_mass
            )
        self.vapor_frac = Reference(
            self.cooler.control_volume.properties_out[:].vapor_frac
        )
        self.split_flow = self._create_split_flow_references()

        # Condensate liquid always is removed first.
        self.phase_separator.split_fraction[:, "vapor_outlet", "Vap"].fix(1.0)
        self.phase_separator.split_fraction[:, "vapor_outlet", "Liq"].fix(0.0)

        # Additional bounds and constraints
        self._additional_constraints()

        # Expand arcs
        pyo.TransformationFactory("network.expand_arcs").apply_to(self)

        self.cooler_to_separator_arc_expanded.flow_mol_equality.deactivate()

        # add an expression for the degree of superheat

        @self.Expression(self.flowsheet().time)
        def degree_of_superheat(self,t):
            return self.splitter.mixed_state[t].temperature - self.splitter.mixed_state[t].temperature_sat

    def _create_split_flow_references(self):
        self.outlet_idx = pyo.Set(initialize=self.outlet_list)
        # Map each (t, o) to the outlet state's flow var
        ref_map = {}
        for o in self.outlet_list:
            if o != "vent":
                outlet_state_block = getattr(self.splitter, f"{o}_state")
                for t in self.flowsheet().time:
                    ref_map[(t, o)] = outlet_state_block[t].flow_mol

        return Reference(ref_map)

    def _additional_constraints(self):
        """
        Additional constraints.
        """

        """
        1) Set lower bounds on flow variables for all external ports.
        """
        [
            state_block[t].flow_mol.setlb(0.0)
            for state_block in (self.inlet_blocks + self.outlet_blocks)
            for t in state_block
        ]

        """
        2) Add overall material balance equation.
        """

        # Write phase-component balances
        @self.Constraint(self.flowsheet().time, doc="Material balance equation")
        def material_balance_equation(b, t):
            pc_set = b.outlet_blocks[0].phase_component_set
            comp_ls = b.outlet_blocks[0].component_list
            phase_ls = b.outlet_blocks[0].phase_list
            return (
                0
                == sum(
                    sum(
                        b.mixer.mixed_state[t].get_material_flow_terms(p, j)
                        - (b.phase_separator.mixed_state[t].get_material_flow_terms(p, j)
                        - b.splitter.vent_state[t].get_material_flow_terms(p, j))
                        for j in comp_ls
                        if (p, j) in pc_set
                    )
                    for p in phase_ls
                )
                - b.balance_flow_mol[t]
            )

        """
        3) Add vent flow balance (scaled by the total known flows).
        """

        @self.Constraint(
            self.flowsheet().time,
            doc="Material balance of known inlet and outlet flows. "
            "Only used for scaling the steam vent min flow, doesn't actually matter.",
        )
        def partial_material_balance_equation(b, t):
            pc_set = b.inlet_blocks[0].phase_component_set
            comp_ls = b.inlet_blocks[0].component_list
            phase_ls = b.inlet_blocks[0].phase_list
            return (
                0
                == sum(
                    sum(
                        sum(
                            o[t].get_material_flow_terms(p, j)
                            for o in b.inlet_blocks[:-1] + b.outlet_blocks[:-2]
                        )
                        for j in comp_ls
                        if (p, j) in pc_set
                    )
                    for p in phase_ls
                )
                - b._partial_total_flow_mol[t]
            )

        eps = 1e-5  # smoothing parameter; smaller = closer to exact max, larger = smoother
        # calculate amount for vent flow: positive balance amount
        @self.Constraint(self.flowsheet().time, doc="Steam vent flow balance.")
        def vent_flow_balance(b, t):
            return 0 == (
                smooth_max(
                    b.balance_flow_mol[t] / (b._partial_total_flow_mol[t] + 1e-6),
                    0.0,
                    eps,
                )
                * (b._partial_total_flow_mol[t] + 1e-6)
                - b.splitter.vent_state[t].flow_mol
            )
        
        # if balance is negative, vent will be zero. makeup will be required.
        @self.Constraint(self.flowsheet().time, doc="Steam makeup balance")
        def makeup_flow_balance(b, t):
            return b.makeup_flow_mol[t] == (
                b.splitter.vent_state[t].flow_mol - b.balance_flow_mol[t]
            )


    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        [getattr(o, "calculate_scaling_factors")() for o in self.unit_ops]

    def initialize_build(self, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialize the Steam Header unit using Sequential Decomposition to determine optimal order
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")

        init_log.info(
            f"Starting {self.parent_block().__class__.__name__} initialization using Sequential Decomposition"
        )
        # Initialize the inverted values
        initialise_inverted(self.cooler, "heat_duty")
        initialise_inverted(self.cooler, "deltaP")

        # create Sequential Decomposition object
        seq = SequentialDecomposition()
        seq.options.select_tear_method = "heuristic"
        seq.options.tear_method = "Wegstein"
        seq.options.iterLim = 1

        # create computation graph
        G = seq.create_graph(self)
        heuristic_tear_set = seq.tear_set_arcs(G, method="heuristic")
        # get calculation order
        order = seq.calculation_order(G)

        for o in heuristic_tear_set:
            print(o.name)

        print("Initialization order:")
        for o in order:
            print(o[0].name)

        # define unit initialisation function
        def init_unit(unit):
            unit.initialize(outlvl=outlvl, **kwargs)

        # run sequential decomposition
        seq.run(self, init_unit)

    def _get_stream_table_contents(self, time_point=0):
        """
        Create stream table showing all inlets, outlets, and liquid outlet
        """
        io_dict = {}

        for inlet_name in self.inlet_list:
            io_dict[inlet_name] = getattr(self, inlet_name)

        for outlet_name in self.outlet_list:
            io_dict[outlet_name] = getattr(self, outlet_name)

        io_dict["condensate_outlet"] = self.condensate_outlet

        return create_stream_table_dataframe(io_dict, time_point=time_point)



# Because there are a lot of things that reference the flowsheet manager, we
# get a circular dependency.
# But this is only a problem for the type hints.
# so instead, we will use this trick:

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .flowsheet_manager import FlowsheetManager
else:
    FlowsheetManager = "FlowsheetManager"
# This only imports the type if we are checking types.
# Then as long as we use the type hints as a forward reference, we are okay.
# https://peps.python.org/pep-0484/#forward-references
# e.g
# from .flowsheet_manager_type import FlowsheetManager
# def __init__(self, flowsheet_manager: "FlowsheetManager"):
# rather than
# from .flowsheet_manager import FlowsheetManager
# def __init__(self, flowsheet_manager: FlowsheetManager):
#


import pyomo.environ as pe# Pyomo environment
from idaes.core import FlowsheetBlock, StateBlock
from idaes.models.unit_models import HeatExchanger
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    PhaseType,
    StateVars,
)
from idaes.models.unit_models.heat_exchanger import HX0DInitializer
from idaes.models.unit_models.heat_exchanger import delta_temperature_lmtd_callback
from idaes.models.properties import iapws95
from ahuora_builder.custom.SimpleEffectivenessHX_DH  import HeatExchangerEffectiveness  
# from milk_props_full_config  import MilkParameterBlock  
from idaes.core.util.model_statistics import degrees_of_freedom
import idaes.logger as idaeslog
from idaes.core.util import DiagnosticsToolbox
from multiprocessing import freeze_support
from property_packages.build_package import build_package

#--------------------------------------------------------------
#Only works with smooth temperature need the Ahuora smooth helmhotlz, the Ahuora smooth Helmholtz does not work on windows
#--------------------------------------------------------------

if __name__ == '__main__':
    freeze_support()
    # Create an empty flowsheet and steam property parameter block.
    model = pe.ConcreteModel()
    model.fs = FlowsheetBlock(dynamic=False)
    model.fs.properties = build_package("helmholtz",["water"],["Liq","Vap"])  #HelmholtzParameterBlock(pure_component="h2o",state_vars=StateVars.PH,phase_presentation=PhaseType.MIX)
    dt = DiagnosticsToolbox(model)
    # Add a Heater model to the flowsheet.
    model.fs.heat_exchanger = HeatExchangerEffectiveness(
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": model.fs.properties},
        tube={"property_package": model.fs.properties},

    )

    model.fs.heat_exchanger.effectiveness.fix(1)
    model.fs.heat_exchanger.shell_inlet.flow_mol.fix(10)
    P_hot = 270280
    T_hot = 200 + 273
    h = model.fs.properties.htpx(T=T_hot*pe.units.K, p= P_hot*pe.units.Pa)
    print(h)
    model.fs.heat_exchanger.shell_inlet.pressure.fix(P_hot)
    model.fs.heat_exchanger.shell_inlet.enth_mol.fix(h)

    P_cold = 101325
    T_cold = 50 + 273
    model.fs.heat_exchanger.tube_inlet.flow_mol.fix(5)
    model.fs.heat_exchanger.tube_inlet.pressure.fix(P_cold)

    h = model.fs.properties.htpx(T=T_cold*pe.units.K, p= P_cold*pe.units.Pa)
    print(h)
    model.fs.heat_exchanger.tube_inlet.enth_mol.fix(h)
    # Perform degrees of freedom analysis
    dof = degrees_of_freedom(model)
    print("Degrees of Freedom:", dof)


    # Initialize the model
    # initializer = HX0DInitializer()
    # initializer.initialize(model.fs.heat_exchanger)
    model.fs.heat_exchanger.initialize(outlvl=idaeslog.INFO_HIGH)
    model.fs.heat_exchanger.report()

    # Solve the models
    solver = pe.SolverFactory('ipopt')
    results = solver.solve(model, tee=True)

    # Display the results
    model.fs.heat_exchanger.report()
    print("Hot inlet enthalpy theory: ", model.fs.properties.htpx(p= P_hot*pe.units.Pa, T= T_hot*pe.units.K), "J/mol")
    print("Cold inlet enthalpy theory: ", model.fs.properties.htpx(p= P_cold*pe.units.Pa, T= T_cold*pe.units.K), "J/mol")
    print("Hot inlet enthalpy: ", pe.value(model.fs.heat_exchanger.hot_side.properties_in[0].enth_mol), "J/mol")
    print("Cold inlet enthalpy: ", pe.value(model.fs.heat_exchanger.cold_side.properties_in[0].enth_mol), "J/mol")
    print("Hot outlet enthalpy_max: ", pe.value(model.fs.heat_exchanger.properties_hotside[0].enth_mol), "J/mol")
    print("Cold outlet enthalpy_max: ", pe.value(model.fs.heat_exchanger.properties_coldside[0].enth_mol), "J/mol")
    print("dH_hot: ", pe.value(model.fs.heat_exchanger.delta_h_hot_qmax[0]), "J/mol")
    print("dH_cold: ", pe.value(model.fs.heat_exchanger.delta_h_cold_qmax[0]), "J/mol")
    print("Duty: ", pe.value(model.fs.heat_exchanger.heat_duty[0])/1e3, "kW")
    print("Qmax: ", pe.value(model.fs.heat_exchanger.Qmax[0])/1e3, "kW")


    eff = pe.value(model.fs.heat_exchanger.effectiveness[0])
    eff_apparent = pe.value(model.fs.heat_exchanger.heat_duty[0]/model.fs.heat_exchanger.Qmax[0])
    print("Effectiveness set: ", eff)
    print("Effectiveness apparent: ", eff_apparent)





### Imports
from pyomo.environ import ConcreteModel, SolverFactory, SolverStatus, TerminationCondition, Block, TransformationFactory, assert_optimal_termination
from pyomo.network import SequentialDecomposition, Port, Arc
from pyomo.core.base.units_container import _PyomoUnit, units 
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import report_statistics, degrees_of_freedom
from idaes.core.util.tables import _get_state_from_port
import idaes.logger as idaeslog
from property_packages.build_package import build_package
from idaes.models.unit_models.pressure_changer import Pump
from idaes.models.unit_models.heater import Heater
from ..power_property_package import PowerParameterBlock
from ..bus import Bus
from pyomo.environ import value

def test_bus():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.power_property_package = PowerParameterBlock()

    m.fs.bus = Bus(
        property_package=m.fs.power_property_package,
        num_inlets=2,
        
        
        )
    m.fs.bus.inlet_1.power.fix(100 * units.kW) 
    m.fs.bus.inlet_2.power.fix(50 * units.kW)
    

    m.fs.bus.eq_power_balance.pprint()

    #m.fs.bus.display()
    
    # Check that the model is fully specified
    assert degrees_of_freedom(m.fs.bus) == 0 

    # check that it solves correctly

    solver = SolverFactory("ipopt")
    solver.solve(m)
    print(*"Post solve power:")

    assert value(m.fs.bus.properties_out[0].power)- 150e3 == 0


#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Base class for unit models
"""

from pyomo.environ import check_optimal_termination
from pyomo.common.config import ConfigValue
from pyomo.network import Port

from idaes.core.base.process_base import (
    declare_process_block_class,
    ProcessBlockData,
    useDefault,
)
from idaes.core.base.property_base import StateBlock
from idaes.core.base.control_volume_base import (
    ControlVolumeBlockData,
    FlowDirection,
    MaterialBalanceType,
)
from idaes.core.util.exceptions import (
    BurntToast,
    ConfigurationError,
    BalanceTypeNotSupportedError,
    InitializationError,
)
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
from idaes.core.util.config import DefaultBool
from idaes.core.initialization import SingleControlVolumeUnitInitializer
from pyomo.environ import Var, units, Expression
from pyomo.dae import Integral


# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("IntegrationBlock")
class IntegrationBlockData(ProcessBlockData):
    """
    This block has no inlets or outlets, but one value that you can get the integral for.
    """

    # Create Class ConfigBlock
    CONFIG = ProcessBlockData.CONFIG()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            default=useDefault,
            domain=DefaultBool,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as a dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )

    def build(self):
        """
        
        """
        super(IntegrationBlockData, self).build()

        self.variable = Var(
            self.flowsheet().time
        )

        #self.time = self.flowsheet().time

        
        
        def _integral(m):
            time_steps = sorted(self.flowsheet().time)
            value = m.variable
            return sum([
                # Trapezium rule
                # average the value at the current and next time step, then multiply by the time difference
                0.5 * (value[time_steps[i]] + value[time_steps[i+1]]) * (time_steps[i+1] - time_steps[i]) 
                for i in range(len(time_steps)-1)
            ])
        
        self.integral = Expression(
            rule=_integral
        )
        
        
        # if self.config.dynamic:
        #     # create a variable to hold the normal value, and one to hold the integral value
        # else:




from idaes.core import declare_process_block_class, MaterialBalanceType
from .water_tank_with_units import WaterTankData
from .add_initial_dynamics import add_initial_dynamics
from pyomo.environ import Reference, Var
from pyomo.dae import DerivativeVar
from pyomo.environ import units as pyunit
from .inverted import add_inverted, initialise_inverted


def CustomTank(*args, **kwargs):
    # In dynamics mode, we need to use the default material balance type of componentPhase.
    # This does balances for liquid and vapor phases separately. This is needed becasue 
    # accumulation terms are phase specific.
    # However, in steady-state, this is not necessary, because we don't have accumulation terms.
    # So this would make the system over-defined, as the state block defines the phase equilibrium already.
    is_dynamic = kwargs.get("dynamic")
    if is_dynamic:
        kwargs["material_balance_type"] = MaterialBalanceType.componentPhase
    else:
        kwargs["material_balance_type"] = MaterialBalanceType.componentTotal
    
    return DynamicTank(*args, **kwargs)
        

@declare_process_block_class("DynamicTank")
class DynamicTankData(WaterTankData):
    """
    Water tank model with dynamic capabilities.
    Some extra properties are added to IDAES's tank model to allow for easier specification of initial conditions.
    """

    def build(self, *args, **kwargs):
        """
        Build method for the DynamicHeaterData class.
        This method initializes the control volume and sets up the model.
        """

        super().build(*args, **kwargs)

        self.flow_rate_out = Reference(self.outlet.flow_mol)

        # add deltaP_inverted as a property
        add_inverted(self,"heat_duty")

        # Because it's hard to specify the initial conditions directly,
        # Create a state block for the initial conditions.

        if not self.config.dynamic:
            return # There is no need to add these extra properties.        

        self.initial_block = self.config.property_package.build_state_block(
            [0],
            defined_state=True,
        )

        if len(self.config.property_package.component_list) > 1:

            # We are assuming that the composition does not change at the initial time step. In theory it could, but
            # we can worry about that later.
            @self.Constraint(
                self.config.property_package.component_list,
                doc="Initial composition constraint",
            )
            def initial_composition_constraint(b, j):
                return (
                    b.initial_block[0].mole_frac_comp[j]
                    == b.control_volume.properties_in[0].mole_frac_comp[j]
                )

        # The initial temperature, pressure, and flow amount is set by the user.
        self.initial_pressure = Var(initialize=101325, units=pyunit.Pa)
        @self.Constraint(doc="Initial pressure constraint")
        def initial_pressure_constraint(b):
            return b.initial_block[0].pressure == b.initial_pressure

        self.initial_holdup = Var(initialize=300, units=pyunit.mol)
        @self.Constraint(doc="Initial flow constraint")
        def initial_holdup_mol_constraint(b):
            return (
                b.initial_block[0].flow_mol * pyunit.s == b.initial_holdup
            )  # cancel out the seconds as we are using it for holdup not accumulation.
        

        self.initial_level = Var(initialize=300, units=pyunit.m)
        @self.Constraint(doc="Initial level constraint")
        def initial_level_constraint(b):
            return b.initial_holdup == b.tank_cross_sect_area * b.initial_level * b.initial_block[0].dens_mol

        self.initial_temperature = Var(initialize=300, units=pyunit.K)
        @self.Constraint(doc="Initial temperature constraint")
        def initial_temperature_constraint(b):
            return b.initial_block[0].temperature == b.initial_temperature

        # The temperature, pressure and flow are used to calculate the other properties.
        @self.Constraint(
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Defining accumulation",
        )
        def initial_material_conditions_constraint(b, p, j):
            return (
                b.initial_block[0].flow_mol
                * b.initial_block[0].mole_frac_phase_comp[p, j]
                + b.control_volume.material_accumulation[0, p, j]
                == b.control_volume.material_holdup[0, p, j]
            )

        @self.Constraint(
            self.config.property_package.phase_list, doc="Defining accumulation"
        )
        def initial_energy_conditions_constraint(b, p):
            return (
                b.initial_block[0].flow_mol
                * b.initial_block[0].phase_frac[p]
                * b.initial_block[0].enth_mol_phase[p]
                + b.control_volume.energy_accumulation[0, p]
                == b.control_volume.energy_holdup[0, p]
            )
        
        

    def initialize(self, *args, **kwargs):
        """
        Initialization method for the DynamicTankData class.
        This method initializes the control volume and sets up the model.
        """
        # Copy initial conditions from inverted properties
        initialise_inverted(self,"heat_duty")

        if self.config.dynamic:
            self.initial_block.initialize()
        
        super().initialize(*args, **kwargs)
